from abc import ABC, abstractmethod
from typing import Any, Literal, Optional, Union

import numpy as np
import numpy.typing as npt

from ..util.seeding import RngType
from .schedulers import NoScheduling, Scheduler


class ExplorationStrategy(ABC):
    """Base class for exploration strategies such as greedy, epsilon-greedy, etc."""

    def __init__(
        self,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
    ) -> None:
        """Instantiates a generic exploration strategy.

        Parameters
        ----------
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies to which callback to hook, i.e., when to step the exploration's
            schedulers (if any) to, e.g., decay the chances of exploring or the
            perturbation strength (see `step` method also). The options are:
             - `on_update` steps the exploration after each agent's update
             - `on_episode_end` steps the exploration after each episode's end
             - `on_timestep_end` steps the exploration after each env's timestep.

            By default, 'on_update' is selected.
        mode : {'gradient-based', 'additive'} optional
            Mode of application of explorative perturbations to the MPC. If `additive`,
            then the drawn pertubation is added to the optimal action computed by the
            MPC. By default, `gradient-based` is selected, and in this mode the
            pertubations enter  directly in the MPC objective and multiplied by the
            first action, thus affecting its gradient.
        """
        super().__init__()
        self._hook = hook
        self._mode = mode

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        """Specifies to which callback to hook, i.e., when to step the exploration's
        schedulers (if any) to, e.g., decay the chances of exploring or the perturbation
        strength (see `step` method also). Can be `None` in case no hook is needed."""
        return self._hook  # override this property if schedulers are used in the class

    @property
    def mode(self) -> Literal["gradient-based", "additive"]:
        """Mode of application of explorative perturbations to the MPC."""
        return self._mode

    @abstractmethod
    def can_explore(self) -> bool:
        """Computes whether, according to the exploration strategy, the agent should
        explore or not at the current instant.

        Returns
        -------
        bool
            `True` if the agent should explore according to this strategy; otherwise,
            `False`.
        """

    @abstractmethod
    def step(self, *args: Any, **kwargs: Any) -> None:
        """Updates the exploration strength and/or probability, in case the strategy
        supports them (usually, by decaying them over time)."""

    @abstractmethod
    def perturbation(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        """Returns a random perturbation."""

    def reset(self, _: RngType = None) -> None:
        """Resets the exploration status, in case it is non-deterministic."""

    def __str__(self) -> str:
        return self.__class__.__name__

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(hook={self.hook},mode={self.mode})"


class NoExploration(ExplorationStrategy):
    """Strategy where no exploration is allowed at any time or, in other words, the
    policy is always deterministic (only based on the current state, and not perturbed).

    This is a special kind of `ExplorationStrategy`, the only one without any
    `hook` and `mode`.
    """

    def __init__(self) -> None:
        """Instiates a no-exploration strategy."""
        super().__init__()
        del self._hook, self._mode

    @property
    def hook(self) -> None:
        """Returns `None`, since no exploration is allowed."""
        return None

    @property
    def mode(self) -> None:
        """Returns no mode."""
        return None

    def can_explore(self) -> bool:
        return False

    def step(self, *_, **__) -> None:
        return

    def perturbation(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        raise NotImplementedError(
            f"Perturbation not implemented in {self.__class__.__name__}"
        )

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"


class GreedyExploration(ExplorationStrategy):
    """Fully greedy strategy for perturbing the policy, thus inducing exploration. This
    strategy always perturbs randomly the policy."""

    def __init__(
        self,
        strength: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
        seed: RngType = None,
    ) -> None:
        """Initializes the greedy exploration strategy.

        Parameters
        ----------
        strength : scheduler or array/supports-algebraic-operations
            The strength of the exploration. If passed in the form of an
            `mpcrl.schedulers.Scheduler`, then the strength can be scheduled to
            decay/increase every time `exploration.step` is called. If an array or
            something other than a scheduler is passed, then this quantity will get
            wrapped in a base scheduler which will kept it constant.
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies to which callback to hook, i.e., when to step the exploration's
            schedulers (if any) to, e.g., decay the chances of exploring or the
            perturbation strength (see `step` method also). The options are:
             - `on_update` steps the exploration after each agent's update
             - `on_episode_end` steps the exploration after each episode's end
             - `on_timestep_end` steps the exploration after each env's timestep.

            By default, 'on_update' is selected.
        mode : {'gradient-based', 'additive'} optional
            Mode of application of explorative perturbations to the MPC. If `additive`,
            then the drawn pertubation is added to the optimal action computed by the
            MPC. By default, `gradient-based` is selected, and in this mode the
            pertubations enter  directly in the MPC objective and multiplied by the
            first action, thus affecting its gradient.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Number to seed the RNG engine used for randomizing the exploration. By
            default, `None`.
        """
        super().__init__(hook, mode)
        if not isinstance(strength, Scheduler):
            strength = NoScheduling[npt.NDArray[np.floating]](strength)
        self.strength_scheduler = strength
        self.reset(seed)

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        # return hook only if the strength scheduler requires to be stepped
        return None if isinstance(self.strength_scheduler, NoScheduling) else self._hook

    def reset(self, seed: RngType = None) -> None:
        """Resets the exploration RNG."""
        self.np_random = np.random.default_rng(seed)

    def can_explore(self) -> bool:
        return True

    def step(self, *_, **__) -> None:
        """Updates the exploration strength according to its scheduler."""
        self.strength_scheduler.step()

    def perturbation(
        self, method: str, *args: Any, **kwargs: Any
    ) -> npt.NDArray[np.floating]:
        """Returns a random perturbation.

        Parameters
        ----------
        method : str
            The name of a method from the ones available to `numpy.random.Generator`,
            e.g., 'random', 'normal', etc.
        args, kwargs
            Args and kwargs with which to call such method.

        Returns
        -------
        array
            An array representing the perturbation.
        """
        return (
            getattr(self.np_random, method)(*args, **kwargs)
            * self.strength_scheduler.value
        )

    def __repr__(self) -> str:
        stn = self.strength_scheduler.value
        return f"{self.__class__.__name__}(stn={stn},hook={self.hook},mode={self.mode})"


class EpsilonGreedyExploration(GreedyExploration):
    """Epsilon-greedy strategy for perturbing the policy, thus inducing exploration.
    This strategy only occasionally perturbs randomly the policy."""

    def __init__(
        self,
        epsilon: Union[Scheduler[float], float],
        strength: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
        seed: RngType = None,
    ) -> None:
        """Initializes the epsilon-greedy exploration strategy.

        Parameters
        ----------
        epsilon : scheduler or float
            The probability to explore. Should be in range [0, 1]. If passed in the form
            of an `mpcrl.schedulers.Scheduler`, then the probability can be scheduled to
            decay/increase every time `exploration.step` is called. If an array or
            something other than a scheduler is passed, then this quantity will get
            wrapped in a base scheduler which will kept it constant.
        strength : scheduler or array/supports-algebraic-operations
            The strength of the exploration. Can be scheduled, see `epsilon`.
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies to which callback to hook, i.e., when to step the exploration's
            schedulers (if any) to, e.g., decay the chances of exploring or the
            perturbation strength (see `step` method also). The options are:
             - `on_update` steps the exploration after each agent's update
             - `on_episode_end` steps the exploration after each episode's end
             - `on_timestep_end` steps the exploration after each env's timestep.

            By default, 'on_update' is selected.
        mode : {'gradient-based', 'additive'} optional
            Mode of application of explorative perturbations to the MPC. If `additive`,
            then the drawn pertubation is added to the optimal action computed by the
            MPC. By default, `gradient-based` is selected, and in this mode the
            pertubations enter  directly in the MPC objective and multiplied by the
            first action, thus affecting its gradient.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Number to seed the RNG engine used for randomizing the exploration. By
            default, `None`.
        """
        super().__init__(strength, hook, mode, seed)
        if not isinstance(epsilon, Scheduler):
            epsilon = NoScheduling[float](epsilon)
        self.epsilon_scheduler = epsilon

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        # return hook only if the strength or epsilon scheduler requires to be stepped
        return (
            None
            if isinstance(self.strength_scheduler, NoScheduling)
            and isinstance(self.epsilon_scheduler, NoScheduling)
            else self._hook
        )

    def can_explore(self) -> bool:
        return self.np_random.random() <= self.epsilon_scheduler.value

    def step(self, *_, **__) -> None:
        """Updates the exploration probability and strength according to their
        schedulers."""
        self.strength_scheduler.step()
        self.epsilon_scheduler.step()

    def __repr__(self) -> str:
        clsn = self.__class__.__name__
        eps = self.epsilon_scheduler.value
        stn = self.strength_scheduler.value
        return f"{clsn}(eps={eps},stn={stn},hook={self.hook},mode={self.mode})"


class OrnsteinUhlenbeckExploration(ExplorationStrategy):
    """
    Exploraiton based on the Ornstein-Uhlenbeck Brownian motion with friction. See
    implementation from  https://github.com/DLR-RM/stable-baselines3/tree/master.

    Note: since this exploration strategy creates a particular noise process, it is
    independent of the agent's `cost_perturbation_method` field.
    """

    def __init__(
        self,
        mean: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        sigma: Union[Scheduler[npt.NDArray[np.floating]], npt.NDArray[np.floating]],
        theta: float = 0.15,
        dt: float = 1.0,
        initial_noise: Optional[npt.ArrayLike] = None,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        mode: Literal["gradient-based", "additive"] = "gradient-based",
        seed: RngType = None,
    ) -> None:
        """Creates a new Ornstein-Uhlenbeck exploration strategy.

        Parameters
        ----------
        mean : scheduler or array/supports-algebraic-operations
            Mean of the stochastic process. Should have the same shape as the action.
        sigma : scheduler or array/supports-algebraic-operations
            Standard deviation of the stochastic process. Should have the same shape as
            the action.
        theta : float, optional
            Coefficient of attraction of the process towards mean, by default `0.15`.
        dt : float, optional
            Time step of the process, by default `1.0`.
        initial_noise : array-like, optional
            A default initial noise. By default `None`, in which case it is set to zero.
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies to which callback to hook, i.e., when to step the exploration's
            schedulers (if any) to, e.g., decay the chances of exploring or the
            perturbation strength (see `step` method also). The options are:
             - `on_update` steps the exploration after each agent's update
             - `on_episode_end` steps the exploration after each episode's end
             - `on_timestep_end` steps the exploration after each env's timestep.

            By default, 'on_update' is selected.
        mode : {'gradient-based', 'additive'} optional
            Mode of application of explorative perturbations to the MPC. If `additive`,
            then the drawn pertubation is added to the optimal action computed by the
            MPC. By default, `gradient-based` is selected, and in this mode the
            pertubations enter  directly in the MPC objective and multiplied by the
            first action, thus affecting its gradient.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Number to seed the RNG engine used for randomizing the exploration. By
            default, `None`.
        """
        super().__init__(hook, mode)
        if not isinstance(mean, Scheduler):
            mean = NoScheduling[npt.NDArray[np.floating]](mean)
        self.mean_scheduler = mean
        if not isinstance(sigma, Scheduler):
            sigma = NoScheduling[npt.NDArray[np.floating]](sigma)
        self.sigma_scheduler = sigma
        self.theta = theta
        self.dt = dt
        self.initial_noise = initial_noise
        self.reset(seed)

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        # return hook only if the mean or sigma scheduler requires to be stepped
        return (
            None
            if isinstance(self.mean_scheduler, NoScheduling)
            and isinstance(self.sigma_scheduler, NoScheduling)
            else self._hook
        )

    def reset(self, seed: RngType = None) -> None:
        """Resets the exploration RNG."""
        self.np_random = np.random.default_rng(seed)
        self._prev_noise = (
            np.zeros_like(self.mean_scheduler.value)
            if self.initial_noise is None
            else np.asarray(self.initial_noise)
        )

    def can_explore(self) -> bool:
        return True

    def step(self, *_, **__) -> None:
        """Updates the mean and std of the noise according to their schedulers."""
        self.mean_scheduler.step()
        self.sigma_scheduler.step()

    def perturbation(self, *_: Any, **__: Any) -> npt.NDArray[np.floating]:
        sigma = self.sigma_scheduler.value
        noise = (
            self._prev_noise
            + (self.theta * self.dt) * (self.mean_scheduler.value - self._prev_noise)
            + np.sqrt(self.dt) * (sigma * self.np_random.normal(size=np.shape(sigma)))
        )
        self._prev_noise = noise
        return noise

    def __repr__(self) -> str:
        clsn = self.__class__.__name__
        mean = self.mean_scheduler.value
        sigma = self.sigma_scheduler.value
        return (
            f"{clsn}(mean={mean},sigma={sigma},theta={self.theta},dt={self.dt},"
            f"hook={self.hook},mode={self.mode})"
        )


class StepWiseExploration(ExplorationStrategy):
    """Wrapper exploration class that enables a base exploration strategy to change only
    every N steps, thus yielding a step-wise strategy with steps of the given length.
    This is useful when, e.g., the exploration strategy must be kept constant across
    time for a number of steps.

    Note
    ----
    This exploration wrapper modifies the exploration chance and magnitude of the
    wrapped base strategy as well as the step behaviour, i.e., the decay of the base
    exploration's schedulers (if any) is enlarged by the step size factor. This is
    because the number of calls to the base exploration's `step` method is reduced by
    a factor of the step size.
    """

    def __init__(
        self,
        base_exploration: ExplorationStrategy,
        step_size: int,
        stepwise_decay: bool = True,
    ) -> None:
        """Creates a step-wise exploration strategy wrapepr.

        Parameters
        ----------
        base_exploration : ExplorationStrategy
            The base exploration strategy to be made step-wise.
        step_size : int
            Size of each step.
        stepwise_decay : bool, optional
            Enables the decay `step` to also be step-wise, i.e., applied only every N
            steps.
        """
        super().__init__()
        del self._hook, self._mode
        self.base_exploration = base_exploration
        self.step_size = step_size
        self._explore_counter = 0
        self._step_counter = 0
        self._stepwise_decay = stepwise_decay

    @property
    def hook(
        self,
    ) -> Optional[Literal["on_update", "on_episode_end", "on_timestep_end"]]:
        """Returns the hook of the base exploration strategy, if any."""
        return self.base_exploration.hook

    @property
    def mode(self) -> Literal["gradient-based", "additive"]:
        """Returns the mode of the base exploration strategy."""
        return self.base_exploration.mode

    def can_explore(self) -> bool:
        # since this method is called at every timestep (when deterministic=False), we
        # decide here if the base exploration is frozen or not, i.e., if we are at the
        # new step or not
        self._explore_counter %= self.step_size
        if self._explore_counter == 0:
            self._cached_can_explore = self._cached_perturbation = None
        self._explore_counter += 1

        if self._cached_can_explore is not None:
            return self._cached_can_explore
        self._cached_can_explore = self.base_exploration.can_explore()
        return self._cached_can_explore

    def step(self, *_, **__) -> None:
        if not self._stepwise_decay:
            return self.base_exploration.step()
        self._step_counter %= self.step_size
        if self._step_counter == 0:
            self.base_exploration.step()
        self._step_counter += 1

    def perturbation(self, *args: Any, **kwargs: Any) -> npt.NDArray[np.floating]:
        if self._cached_perturbation is not None:
            return self._cached_perturbation
        self._cached_perturbation = self.base_exploration.perturbation(*args, **kwargs)
        return self._cached_perturbation

    def __repr__(self) -> str:
        clsn = self.__class__.__name__
        bclsn = self.base_exploration.__class__.__name__
        h = self.hook
        m = self.mode
        return f"{clsn}(base={bclsn},step_size={self.step_size},hook={h},mode={m})"



from collections.abc import Collection, Iterable, Iterator
from itertools import chain
from typing import Any, Generic, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from csnlp import Solution, wrappers
from csnlp.core.cache import invalidate_caches_of
from csnlp.util.io import SupportsDeepcopyAndPickle
from csnlp.wrappers import Mpc
from gymnasium import Env
from gymnasium.spaces import Box
from typing_extensions import TypeAlias

from ...core.callbacks import AgentCallbackMixin
from ...core.exploration import ExplorationStrategy, NoExploration
from ...core.warmstart import WarmStartStrategy
from ...util.named import Named
from ...util.seeding import RngType, mk_seed

SymType = TypeVar("SymType", cs.SX, cs.MX)
ActType: TypeAlias = Union[npt.ArrayLike, dict[str, npt.ArrayLike]]
ObsType: TypeAlias = Union[npt.ArrayLike, dict[str, npt.ArrayLike]]


def _update_dicts(sinks: Iterable[dict], source: dict) -> Iterator[dict]:
    """Internal utility for updating dicts `sinks` with one `source`."""
    for sink in sinks:
        sink.update(source)
        yield sink


class Agent(Named, SupportsDeepcopyAndPickle, AgentCallbackMixin, Generic[SymType]):
    """Simple MPC-based agent with a fixed (i.e., non-learnable) MPC controller.

    In this agent, the MPC controller is used as policy provider, as well as to provide
    the value function `V(s)` and quality function `Q(s,a)`, where `s` and `a` are the
    state and action of the environment, respectively. However, this class does not use
    any RL method to improve its MPC policy."""

    cost_perturbation_method = "normal"
    cost_perturbation_parameter = "cost_perturbation"
    init_action_parameter = init_action_constraint = "a_init"

    def __init__(
        self,
        mpc: Mpc[SymType],
        fixed_parameters: Union[
            None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]
        ] = None,
        exploration: Optional[ExplorationStrategy] = None,
        warmstart: Union[
            Literal["last", "last-successful"], WarmStartStrategy
        ] = "last-successful",
        use_last_action_on_fail: bool = False,
        remove_bounds_on_initial_action: bool = False,
        name: Optional[str] = None,
    ) -> None:
        """Instantiates an agent with an MPC controller.

        Parameters
        ----------
        mpc : Mpc[casadi.SX or MX]
            The MPC controller used as policy provider by this agent. The instance is
            modified in place to create the approximations of the state function `V(s)`
            and action value function `Q(s,a)`, so it is recommended not to modify it
            further after initialization of the agent. Moreover, some parameter and
            constraint names will need to be created, so an error is thrown if these
            names are already in use in the mpc. These names are under the attributes
            `perturbation_parameter`, `action_parameter` and `action_constraint`.
        fixed_parameters : dict[str, array_like] or collection of, optional
            A dict (or collection of dict, in case of `csnlp.MultistartNlp`) whose keys
            are the names of the MPC parameters and the values are their corresponding
            values. Use this to specify fixed parameters, that is, non-learnable. If
            `None`, then no fixed parameter is assumed.
        exploration : ExplorationStrategy, optional
            Exploration strategy for inducing exploration in the online MPC policy. By
            default `None`, in which case `NoExploration` is used.
        warmstart: "last" or "last-successful" or WarmStartStrategy, optional
            The warmstart strategy for the MPC's NLP. If `last-successful`, the last
            successful solution is used to warm start the solver for the next iteration.
            If `last`, the last solution is used, regardless of success or failure.
            Furthermore, a `WarmStartStrategy` object can be passed to specify a
            strategy for generating multiple warmstart points for the NLP. This is
            useful to generate multiple initial conditions for very non-convex problems.
            Can only be used with an MPC that has an underlying multistart NLP problem
            (see `csnlp.MultistartNlp`).
        use_last_action_on_fail : bool, optional
            In case the MPC solver fails
             * if `False`, the action from the last solver's iteration is returned
               anyway (though suboptimal)
             * if `True`, the action from the last successful call to the MPC is
               returned instead (if the MPC has been solved at least once successfully).

            By default, `False`.
        remove_bounds_on_initial_action : bool, optional
            When `True`, the upper and lower bounds on the initial action are removed in
            the action-value function approximator Q(s,a) since the first action is
            constrained to be equal to the initial action. This is useful to avoid
            issues in the LICQ of the NLP. However, it can lead to numerical problems.
            By default, `False`.
        name : str, optional
            Name of the agent. If `None`, one is automatically created from a counter of
            the class' instancies.

        Raises
        ------
        ValueError
            Raises if
             * the given mpc has no control action as optimization variable
             * the reserved parameter and constraint names are already in use
             * a multistart MPC is given, but the warmstart strategy asks for an
               incompatible number of starting points to be generated
             * a warmstart strategy is given, but the MPC does not have an underlying
               multistart NLP problem, so it cannot handle multiple starting points.
        """
        if isinstance(warmstart, str):
            warmstart = WarmStartStrategy(warmstart)
        ws_points = warmstart.n_points
        if mpc.is_multi and ws_points != 0 and mpc.nlp.starts - ws_points not in (0, 1):
            raise ValueError(
                "A multistart MPC was given with {mpc.nlp.starts} multistarts, but the "
                f"given warmstart strategy asks for {ws_points} starting points. "
                "Expected either 0 warmstart points (i.e., it is disabled), or the same"
                " number as MPC's multistarts, or at most one less."
            )
        elif not mpc.is_multi and ws_points > 0:
            raise ValueError(
                "Got a warmstart strategy with more than 0 starting points, but the "
                "given does not have an underlying multistart NLP problem."
            )
        Named.__init__(self, name)
        SupportsDeepcopyAndPickle.__init__(self)
        AgentCallbackMixin.__init__(self)
        self._fixed_pars = fixed_parameters
        if exploration is None:
            exploration = NoExploration()
        self._exploration = exploration
        self._warmstart = warmstart
        self._last_action_on_fail = use_last_action_on_fail
        self._last_solution: Optional[Solution[SymType]] = None
        self._last_action: Optional[cs.DM] = None
        self._V, self._Q = self._setup_V_and_Q(mpc, remove_bounds_on_initial_action)
        self._post_setup_V_and_Q()

    @property
    def unwrapped(self) -> "Agent":
        """Gets the underlying wrapped instance of an agent."""
        return self

    def is_wrapped(self, *args: Any, **kwargs: Any) -> bool:
        """Gets whether the agent instance is wrapped or not by the wrapper type."""
        return False

    @property
    def V(self) -> Mpc[SymType]:
        """Gets the MPC function approximation of the state value function `V(s)`."""
        return self._V

    @property
    def Q(self) -> Mpc[SymType]:
        """Gets the MPC function approximation of the action value function `Q(s,a)`."""
        return self._Q

    @property
    def fixed_parameters(
        self,
    ) -> Union[None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]]:
        """Gets the fixed parameters of the MPC controller (i.e., non-learnable). Can be
        an iterable in the case of `csnlp.MultistartNpl`, where multiple parameters can
        be specified, one for each scenario in the MPC scheme."""
        return self._fixed_pars

    @property
    def exploration(self) -> ExplorationStrategy:
        """Gets the exploration strategy used within this agent."""
        return self._exploration

    @property
    def warmstart(self) -> WarmStartStrategy:
        """Gets the warm start strategy used within this agent."""
        return self._warmstart

    def reset(self, seed: RngType = None) -> None:
        """Resets the agent's internal variables and exploration's RNG."""
        self._last_solution = None
        self._last_action = None
        self.warmstart.reset(seed)
        self.exploration.reset(seed)

    def solve_mpc(
        self,
        mpc: Mpc[SymType],
        state: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        action: Union[None, npt.ArrayLike, dict[str, npt.ArrayLike]] = None,
        perturbation: Optional[npt.ArrayLike] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        store_solution: bool = True,
    ) -> Solution[SymType]:
        """Solves the agent's specific MPC optimal control problem.

        Parameters
        ----------
        mpc : Mpc
            The MPC problem to solve, either `Agent.V` or `Agent.Q`.
        state : array_like or dict[str, array_like]
            A 1D array representing the value of all states of the MPC, concatenated.
            Otherwise, a dict whose keys are the names of each state, and values are
            their numerical values.
        action : array_like or dict[str, array_like], optional
            Same for `state`, for the action. Only valid if evaluating the action value
            function `Q(s,a)`. For this reason, it can be `None` for `V(s)`.
        perturbation : array_like, optional
            The gradient-based cost perturbation used to induce exploration in `V(s)`.
            Should be `None` for `Q(s,a)`, or in case of other types of exploration.
        vals0 : dict[str, array_like] or iterable of, optional
            A dict (or an iterable of dict, in case of `csnlp.MultistartNlp`), whose
            keys are the names of the MPC variables, and values are the numerical
            initial values of each variable. Use this to warm-start the MPC. If `None`,
            and a previous solution (possibly, successful) is available, the MPC solver
            is automatically warm-started. If an iterable is passed instead, the
            warm-starting strategy is bypassed.
        store_solution : bool, optional
            By default, the mpc solution is stored accordingly to the `warmstart`
            strategy. If set to `False`, this flag allows to disable the behaviour for
            this particular solution.

        Returns
        -------
        Solution
            The solution of the MPC.
        """
        # convert state keys into initial state keys (i.e., with "_0")
        if isinstance(state, dict):
            x0_dict = {f"{k}_0": v for k, v in state.items()}
        else:
            mpcstates = mpc.initial_states
            if len(mpcstates) == 1:
                states = (state,)
            else:
                cumsizes = np.cumsum([s.shape[0] for s in mpcstates.values()][:-1])
                states = np.split(state, cumsizes)
            x0_dict = dict(zip(mpcstates.keys(), states))

        # convert action dict to vector if not None
        if action is None:
            u0_vec = None
        elif isinstance(action, dict):
            u0_vec = cs.vertcat(*(action[k] for k in mpc.actions.keys()))
        else:
            u0_vec = action

        # merge (initial) state, action and perturbation in unique dict
        additional_pars = x0_dict
        if u0_vec is not None:
            additional_pars[self.init_action_parameter] = u0_vec
        if self.cost_perturbation_parameter in mpc.parameters:
            additional_pars[self.cost_perturbation_parameter] = (
                0 if perturbation is None else perturbation
            )

        # create pars and vals0
        pars = self._get_parameters()
        if pars is None:
            pars = additional_pars
        elif isinstance(pars, dict):
            pars.update(additional_pars)
        else:  # iterable of dict
            pars = _update_dicts(pars, additional_pars)
        if vals0 is None and self._last_solution is not None:
            vals0 = self._last_solution.vals

        # use the warmstart strategy to generate multiple initial points for the NLP if
        # the NLP supports multi and `vals0` is not already an iterable of dict
        if mpc.is_multi and (vals0 is None or isinstance(vals0, dict)):
            more_vals0s = self._warmstart.generate(vals0)
            if mpc.nlp.starts > self._warmstart.n_points:
                # the difference between these two has been checked to be at most one,
                # meaning we can include `vals0` itself
                more_vals0s = chain((vals0,), more_vals0s)
            vals0 = more_vals0s

        # solve and store solution
        sol = mpc(pars, vals0)
        if store_solution and (self._warmstart.store_always or sol.success):
            self._last_solution = sol
        return sol

    def state_value(
        self,
        state: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        deterministic: bool = False,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        action_space: Optional[Box] = None,
        **kwargs,
    ) -> tuple[cs.DM, Solution[SymType]]:
        """Computes the state value function `V(s)` approximated by the MPC.

        Parameters
        ----------
        state : array_like or dict[str, array_like]
            The state `s` at which to evaluate the value function `V(s)`. Can either be
            a dict of state names and values, or a single concatenated column vector of
            all the states.
        deterministic : bool, optional
            If `True`, the cost of the MPC is perturbed according to the exploration
            strategy to induce some exploratory behaviour. Otherwise, no perturbation is
            performed. By default, `deterministic=False`.
        vals0 : dict[str, array_like] or iterable of, optional
            A dict (or an iterable of dict, in case of `csnlp.MultistartNlp`), whose
            keys are the names of the MPC variables, and values are the numerical
            initial values of each variable. Use this to warm-start the MPC. If `None`,
            and a previous solution (possibly, successful) is available, the MPC solver
            is automatically warm-started.
        action_space : gymnasium.spaces.Box, optional
            The action space of the environment. If not `None`, it is used in case an
            additive exploration perturbation is summed to the action in order to clip
            it back into the action space.

        Returns
        -------
        casadi.DM
            The first optimal action according to the solution of `V(s)`, possibly
            perturbed by exploration noise
        Solution
            The solution of the MPC approximating `V(s)` at the given state.
        """
        V = self._V
        exploration = self._exploration
        exploration_mode = exploration.mode
        na = V.na
        if deterministic or exploration_mode is None or not exploration.can_explore():
            pert = None
        else:
            pert = exploration.perturbation(self.cost_perturbation_method, size=(na, 1))

        grad_pert = pert if exploration_mode == "gradient-based" else None
        sol = self.solve_mpc(V, state, perturbation=grad_pert, vals0=vals0, **kwargs)
        action_opt = cs.vertcat(*(sol.vals[u][:, 0] for u in V.actions.keys()))

        if sol.success:
            self._last_action = action_opt
        elif self._last_action_on_fail and self._last_action is not None:
            action_opt = self._last_action

        if pert is not None and exploration_mode == "additive":
            action_opt_noisy = action_opt + pert
            if action_space is not None:
                lb = action_space.low.reshape(na, 1)
                ub = action_space.high.reshape(na, 1)
                action_opt_noisy = np.clip(action_opt_noisy, lb, ub)
                # if np.equal(action_opt_noisy, action_opt).all():
                #     action_opt_noisy = np.clip(action_opt - pert, lb, ub)
            action_opt = action_opt_noisy
        return action_opt, sol

    def action_value(
        self,
        state: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        action: Union[npt.ArrayLike, dict[str, npt.ArrayLike]],
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        **kwargs,
    ) -> Solution[SymType]:
        """Computes the action value function `Q(s,a)` approximated by the MPC.

        Parameters
        ----------
        state : array_like or dict[str, array_like]
            The state `s` at which to evaluate the value function `Q(s,a)`. Can either
            be a dict of state names and values, or a single concatenated column vector
            of all the states.
        action : array_like or dict[str, array_like]
            The action `a` at which to evaluate the value function `Q(s,a)`. Can either
            be a dict of action names and values, or a single concatenated column vector
            of all the actions.
        vals0 : dict[str, array_like] or iterable of, optional
            A dict (or an iterable of dict, in case of `csnlp.MultistartNlp`), whose
            keys are the names of the MPC variables, and values are the numerical
            initial values of each variable. Use this to warm-start the MPC. If `None`,
            and a previous solution (possibly, successful) is available, the MPC solver
            is automatically warm-started.

        Returns
        -------
        Solution
            The solution of the MPC approximating `Q(s,a)` at given state and action.
        """
        return self.solve_mpc(self._Q, state, action, vals0=vals0, **kwargs)

    def evaluate(
        self,
        env: Env[ObsType, ActType],
        episodes: int,
        deterministic: bool = True,
        seed: RngType = None,
        raises: bool = True,
        env_reset_options: Optional[dict[str, Any]] = None,
    ) -> npt.NDArray[np.floating]:
        """Evaluates the agent in a given environment.

        Note: after solving `V(s)` for the current state `s`, the action is computed and
        passed to the environment as the concatenation of the first optimal action
        variables of the MPC (see `csnlp.Mpc.actions`).

        Parameters
        ----------
        env : Env[ObsType, ActType]
            A gym environment where to test the agent in.
        episodes : int
            Number of evaluation episodes.
        deterministic : bool, optional
            Whether the agent should act deterministically; by default, `True`.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Agent's and each env's RNG seed.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.
        env_reset_options : dict, optional
            Additional information to specify how the environment is reset at each
            evalution episode (optional, depending on the specific environment).

        Returns
        -------
        array of doubles
            The cumulative returns (one return per evaluation episode).

        Raises
        ------
            Raises if the MPC optimization solver fails and `warns_on_exception=False`.
        """
        rng = np.random.default_rng(seed)
        self.reset(rng)
        returns = np.zeros(episodes)
        self.on_validation_start(env)

        for episode in range(episodes):
            state, _ = env.reset(seed=mk_seed(rng), options=env_reset_options)
            truncated, terminated, timestep = False, False, 0
            self.on_episode_start(env, episode, state)

            while not (truncated or terminated):
                action, sol = self.state_value(state, deterministic)
                if not sol.success:
                    self.on_mpc_failure(episode, timestep, sol.status, raises)

                state, r, truncated, terminated, _ = env.step(action)
                self.on_env_step(env, episode, timestep)

                returns[episode] += r
                timestep += 1
                self.on_timestep_end(env, episode, timestep)

            self.on_episode_end(env, episode, returns[episode])

        self.on_validation_end(env, returns)
        return returns

    def _setup_V_and_Q(
        self, mpc: Mpc[SymType], remove_bounds_on_initial_action: bool
    ) -> tuple[Mpc[SymType], Mpc[SymType]]:
        """Internal utility to setup the function approximators for the value function
        V(s) and the quality function Q(s,a)."""
        na = mpc.na
        if na <= 0:
            raise ValueError(f"Expected Mpc with na>0; got na={na} instead.")

        # create V and Q function approximations
        V, Q = mpc, mpc.copy()
        V.unwrapped.name += "_V"
        Q.unwrapped.name += "_Q"

        # for Q, add the additional constraint on the initial action to be equal to a0,
        # and remove the now useless upper/lower bounds on the initial action
        a0 = Q.nlp.parameter(self.init_action_parameter, (na, 1))
        u0 = cs.vcat(mpc.first_actions.values())
        Q.nlp.constraint(self.init_action_constraint, u0, "==", a0)
        if remove_bounds_on_initial_action:
            for name, a in mpc.first_actions.items():
                na_ = a.size1()
                Q.nlp.remove_variable_bounds(name, "both", ((r, 0) for r in range(na_)))

        # for V, add the cost perturbation parameter (only if gradient-based)
        if self._exploration.mode == "gradient-based":
            perturbation = V.nlp.parameter(self.cost_perturbation_parameter, (na, 1))
            f = V.nlp.f
            if mpc.is_wrapped(wrappers.NlpScaling):
                f = mpc.scale(f)
            V.nlp.minimize(f + cs.dot(perturbation, u0))

        # invalidate caches for V and Q since some modifications have been done
        for nlp in (V, Q):
            nlp_ = nlp
            while nlp_ is not nlp_.unwrapped:
                invalidate_caches_of(nlp_)
                nlp_ = nlp_.nlp
            invalidate_caches_of(nlp_.unwrapped)
        return V, Q

    def _post_setup_V_and_Q(self) -> None:
        """Internal utility that is run after the creation of V and Q, allowing for
        further customization in inheriting classes."""
        # warn user of any constraints that linearly includes x0 and u0 (aside from
        # x(0)==s0 and u(0)==a0), which may thus lead to LICQ-failure
        # - u0 in Q should only appear in the 1st dynamics constraint and a0 con
        # - u0 in V should only appear in the 1st dynamics constraint and lbx/ubx
        # - x0 in V, Q should only appear in the 1st dynamics constraint and s0 con
        # for mpc in (self._V, self._Q):
        #     name = mpc.unwrapped.name[-1]
        #     u0 = cs.vcat(self._V.first_actions.values())
        #     x0 = cs.vvcat(self._V.first_states.values())
        #     con = cs.vertcat(mpc.g, mpc.h, mpc.h_lbx, mpc.h_ubx)

        #     if mpc.unwrapped.sym_type.__name__ == "SX":
        #         nnz_con_u0 = len(set(cs.jacobian_sparsity(con, u0).get_triplet()[0]))
        #         nnz_con_x0 = len(set(cs.jacobian_sparsity(con, x0).get_triplet()[0]))
        #     else:
        #         nnz_con_u0 = len(  # computes the same as above, but for MX
        #             set(
        #                 chain.from_iterable(
        #                     cs.jacobian(con, a)[:, : a.size1()]
        #                     .sparsity()
        #                     .get_triplet()[0]
        #                     for a in mpc.actions.values()
        #                 )
        #             )
        #         )
        #         nnz_con_x0 = len(
        #             set(
        #                 chain.from_iterable(
        #                     cs.jacobian(con, s)[:, : s.size1()]
        #                     .sparsity()
        #                     .get_triplet()[0]
        #                     for s in mpc.states.values()
        #                 )
        #             )
        #         )

        #     nnz_exp_u0 = mpc.ns + (mpc.na * 2 if name == "V" else mpc.na)
        #     if nnz_con_u0 > nnz_exp_u0:
        #         warnings.warn(
        #             f"detected {nnz_con_u0} (expected {nnz_exp_u0}) constraints on "
        #             f"initial actions in {name}; make sure that the initial action is"
        #             "not overconstrained (LICQ may be compromised).",
        #             RuntimeWarning,
        #         )
        #     nnz_exp_x0 = mpc.ns * 2
        #     if nnz_con_x0 > nnz_exp_x0:
        #         warnings.warn(
        #             f"detected {nnz_con_x0} (expected {nnz_exp_x0}) constraints on "
        #             f"initial states in {name}; make sure that the initial state is "
        #             "not overconstrained (LICQ may be compromised).",
        #             RuntimeWarning,
        #         )

    def _get_parameters(
        self,
    ) -> Union[None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]]:
        """Internal utility to retrieve parameters of the MPC in order to solve it.
        `Agent` has no learnable parameter, so only fixed parameters are returned."""
        return self._fixed_pars

    def __deepcopy__(self, memo: Optional[dict[int, list[Any]]] = None) -> "Agent":
        """Ensures that the copy has a new name."""
        y = super().__deepcopy__(memo)
        if hasattr(y, "name"):
            y.name += "_copy"
        return y



from abc import ABC, abstractmethod
from collections.abc import Iterable
from itertools import accumulate, repeat
from operator import mul
from typing import Generic, TypeVar, Union

import numpy as np

ScType = TypeVar("ScType")
ScType.__doc__ = "A type that supports basic algebraic operations."


class Scheduler(ABC, Generic[ScType]):
    """Schedulers are helpful classes to update or decay different quantities, such as
    learning rates and/or exploration probability.

    Note
    ----
    If the scheduler has a final iteration, it is expected to raise a `StopIteration`
    exception when the last iteration is reached.
    """

    def __init__(self, init_value: ScType) -> None:
        """Builds the scheduler.

        Parameters
        ----------
        init_value : supports-algebraic-operations
            Initial value that will be updated by this scheduler.
        """
        super().__init__()
        self.value = init_value

    @abstractmethod
    def step(self) -> None:
        """Updates the value of the scheduler by one step.

        Raises
        ------
        StopIteration
            Raises if the final iteration of the scheduler (if any) has been reached and
            `step` was called again.
        """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.value})"

    def __str__(self) -> str:
        return self.__repr__()


class NoScheduling(Scheduler[ScType]):
    """Scheduler that actually performs no scheduling and holds the initial value
    constant."""

    def step(self) -> None:
        return


class ExponentialScheduler(Scheduler[ScType]):
    """Exponentiallly decays the value of the scheduler by `factor` every step, i.e.,
    after k steps the value `value_k = init_value * factor**k`."""

    def __init__(self, init_value: ScType, factor: ScType) -> None:
        """Builds the exponential scheduler.

        Parameters
        ----------
        init_value : supports-algebraic-operations
            Initial value that will be updated by this scheduler.
        factor : Tsc
            The exponential factor to decay the initial value with.
        """
        super().__init__(init_value)
        self.factor = factor

    def step(self) -> None:
        self.value *= self.factor  # type: ignore[operator]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(x={self.value},factor={self.factor})"


class LinearScheduler(Scheduler[ScType]):
    """Linearly updates the initial value of the scheduler towards a final one to be
    reached in N total steps, i.e., after k steps, the value is
    `value_k = init_value + (final_value - init_value) * k / total_steps`.
    """

    def __init__(
        self, init_value: ScType, final_value: ScType, total_steps: int
    ) -> None:
        """Builds the exponential scheduler.

        Parameters
        ----------
        init_value : supports-algebraic-operations
            Initial value that will be updated by this scheduler.
        final_value : supports-algebraic-operations
            Final value that will be reached by the scheduler after `total_steps`.
        total_steps : int
            Total number of steps to linearly interpolate between `init_value` and
            `final_value`.
        """
        super().__init__(init_value)
        increment = (final_value - init_value) / total_steps  # type: ignore[operator]
        self.init_value = init_value
        self.final_value = final_value
        self.total_steps = total_steps
        self.generator = accumulate(repeat(increment, total_steps), initial=init_value)
        next(self.generator)

    def step(self) -> None:
        self.value = next(self.generator)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.value},x0={self.init_value},"
            f"xf={self.final_value},N={self.total_steps})"
        )


class LogLinearScheduler(ExponentialScheduler[ScType]):
    """Updates the initial value of the scheduler towards a final one to be
    reached in N total steps in a linear fashion between the exponents of the two, i.e.,
    after k steps, the value is
    `value_k = init_value * exp(ln(final_value / init_value) / total_steps)**k`.
    """

    def __init__(
        self, init_value: ScType, final_value: ScType, total_steps: int
    ) -> None:
        factor = np.exp(
            np.log(final_value / init_value) / total_steps  # type: ignore[operator]
        )
        super().__init__(init_value, factor)
        self.init_value = init_value
        self.final_value = final_value
        self.total_steps = total_steps
        self.generator = accumulate(
            repeat(factor, total_steps), mul, initial=init_value
        )
        next(self.generator)

    def step(self) -> None:
        self.value = next(self.generator)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(x={self.value},x0={self.init_value},"
            f"xf={self.final_value},N={self.total_steps})"
        )


class Chain(Scheduler[ScType]):
    """Chains multiple schedulers together."""

    def __init__(
        self,
        schedulers: Iterable[Union[Scheduler[ScType], tuple[Scheduler[ScType], int]]],
    ) -> None:
        """Builds the chain of schedulers.

        Parameters
        ----------
        schedulers : iterable of schedulers or (scheduler, int)
            An iterable of schedulers to chain together. If a tuple is passed, the first
            element is expected to be a scheduler, and the second one an integer
            indicating the number of steps to run that scheduler for. This is useful in
            case the scheduler has no fixed number of steps, such as `LinearScheduler`.
        """
        self.schedulers = iter(schedulers)
        self._next_scheduler()
        super().__init__(self._current_scheduler.value)

    def step(self) -> None:
        if self._current_steps is not None:
            self._current_scheduler.step()
            self._current_steps -= 1
            if self._current_steps == 0:
                self._next_scheduler()
        else:
            try:
                self._current_scheduler.step()
            except StopIteration:
                self._next_scheduler()
        self.value = self._current_scheduler.value

    def _next_scheduler(self) -> None:
        """Fetches the next scheduler in the chain. If the chain is over, raises
        `StopIteration`, which is in line with the behaviour of other schedulers."""
        scheduler = next(self.schedulers)
        if isinstance(scheduler, tuple):
            self._current_scheduler, self._current_steps = scheduler
        else:
            self._current_scheduler = scheduler
            self._current_steps = None

    def __repr__(self) -> str:
        s = self._current_scheduler
        return f"{self.__class__.__name__}(x={self.value},current={s})"



from typing import Literal, Optional, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.parameters import LearnableParametersDict, SymType
from ..core.schedulers import Scheduler
from .gradient_based_optimizer import GradientBasedOptimizer, LrType


class Adam(GradientBasedOptimizer[SymType, LrType]):
    """Adam and AdamW optimizers, based on [1,2] and [3,4], respectively. AMSGrad is
    also supported [5].

    References
    ----------
    [1] Kingma, D.P. and Ba, J., 2014. Adam: A method for stochastic optimization.
        arXiv preprint arXiv:1412.6980.
    [2] Loshchilov, I. and Hutter, F., 2017. Decoupled weight decay regularization.
        arXiv preprint arXiv:1711.05101.
    [3] Adam - PyTorch 2.1 documentation.
        https://pytorch.org/docs/stable/generated/torch.optim.Adam.html
    [4] AdamW - PyTorch 2.1 documentation.
        https://pytorch.org/docs/stable/generated/torch.optim.AdamW.html
    [5] Reddi, S.J., Kale, S. and Kumar, S., 2019. On the convergence of adam and
        beyond. arXiv preprint arXiv:1904.09237.
    """

    _order = 1
    _hessian_sparsity = "diag"
    """In Adam, hessian is at most diagonal, i.e., in case we have constraints."""

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        betas: tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0.0,
        decoupled_weight_decay: bool = False,
        amsgrad: bool = False,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
    ) -> None:
        """Instantiates the optimizer.

        Parameters
        ----------
        learning_rate : float/array, scheduler
            The learning rate of the optimizer. A float/array can be passed in case the
            learning rate must stay constant; otherwise, a scheduler can be passed which
            will be stepped `on_update` by default (see `hook` argument).
        betas : tuple of 2 floats, optional
            Coefficients used for computing running averages of gradient and its square.
            By default, they are set to `(0.9, 0.999)`.
        eps : float, optional
            Term added to the denominator to improve numerical stability. By default, it
            is set to `1e-8`.
        weight_decay : float, optional
            A positive float that specifies the decay of the learnable parameters in the
            form of an L2 regularization term. By default, it is set to `0.0`, so no
            decay/regularization takes place.
        decoupled_weight_decay : bool, optional
            If `False`, the optimizer is Adam. Otherwise, it is `AdamW`. By default, it
            is `False`.
        amsgrad : bool, optional
            If `True`, uses the AMSGrad variant. By default, it is `False`.
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies when to step the optimizer's learning rate's scheduler to decay
            its value (see `step` method also). This allows to vary the rate over the
            learning iterations. The options are:
             - `on_update` steps the learning rate after each agent's update
             - `on_episode_end` steps the learning rate after each episode's end
             - `on_timestep_end` steps the learning rate after each env's timestep.

            By default, 'on_update' is selected.
        max_percentage_update : float, optional
            A positive float that specifies the maximum percentage change the learnable
            parameters can experience in each update. For example,
            `max_percentage_update=0.5` means that the parameters can be updated by up
            to 50% of their current value. By default, it is set to `+inf`. If
            specified, the update becomes constrained and has to be solved as a QP,
            which is inevitably slower than its unconstrained counterpart.
        """
        super().__init__(learning_rate, hook, max_percentage_update)
        self.weight_decay = weight_decay
        self.beta1, self.beta2 = betas
        self.eps = eps
        self.decoupled_weight_decay = decoupled_weight_decay
        self.amsgrad = amsgrad

    def set_learnable_parameters(self, pars: LearnableParametersDict[SymType]) -> None:
        super().set_learnable_parameters(pars)
        # initialize also running averages
        n = pars.size
        self._exp_avg = np.zeros(n, dtype=float)
        self._exp_avg_sq = np.zeros(n, dtype=float)
        self._max_exp_avg_sq = np.zeros(n, dtype=float) if self.amsgrad else None
        self._step = 0

    def _first_order_update(
        self, gradient: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        theta = self.learnable_parameters.value

        # compute candidate update
        weight_decay = self.weight_decay
        lr = self.lr_scheduler.value
        if weight_decay > 0.0:
            if self.decoupled_weight_decay:  # i.e., AdamW
                theta = theta * (1 - weight_decay * lr)
            else:
                gradient = gradient + weight_decay * theta
        self._step += 1
        dtheta, self._exp_avg, self._exp_avg_sq, self._max_exp_avg_sq = _adam(
            self._step,
            gradient,
            self._exp_avg,
            self._exp_avg_sq,
            lr,
            self.beta1,
            self.beta2,
            self.eps,
            self._max_exp_avg_sq,
        )

        # if unconstrained, apply the update directly; otherwise, solve the QP
        solver = self._update_solver
        if solver is None:
            return theta + dtheta, None
        lbx, ubx = self._get_update_bounds(theta)
        sol = solver(h=cs.DM.eye(theta.shape[0]), g=-dtheta, lbx=lbx, ubx=ubx)
        dtheta = np.asarray(sol["x"].elements())
        stats = solver.stats()
        return theta + dtheta, None if stats["success"] else stats["return_status"]


def _adam(
    step: int,
    g: np.ndarray,
    exp_avg: np.ndarray,
    exp_avg_sq: np.ndarray,
    lr: LrType,
    beta1: float,
    beta2: float,
    eps: float,
    max_exp_avg_sq: Optional[np.ndarray],
) -> tuple[np.ndarray, np.ndarray, np.ndarray, Optional[np.ndarray]]:
    """Computes the update's change according to Adam algorithm."""
    exp_avg = beta1 * exp_avg + (1 - beta1) * g
    exp_avg_sq = beta2 * exp_avg_sq + (1 - beta2) * np.square(g)

    bias_correction1 = 1 - beta1**step
    bias_correction2 = 1 - beta2**step
    step_size = lr / bias_correction1
    bias_correction2_sqrt = np.sqrt(bias_correction2)

    if max_exp_avg_sq is None:
        denom = np.sqrt(exp_avg_sq) / bias_correction2_sqrt + eps
    else:  # i.e., AMSGrad
        max_exp_avg_sq = np.maximum(max_exp_avg_sq, exp_avg_sq)
        denom = np.sqrt(max_exp_avg_sq) / bias_correction2_sqrt + eps
    dtheta = -step_size * (exp_avg / denom)
    return dtheta, exp_avg, exp_avg_sq, max_exp_avg_sq



from collections.abc import Collection, Iterable
from typing import Callable, Generic, Literal, Optional, SupportsFloat, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from csnlp import Solution
from csnlp.wrappers import Mpc, NlpSensitivity
from gymnasium import Env
from typing_extensions import TypeAlias

from ..core.experience import ExperienceReplay
from ..core.exploration import ExplorationStrategy, NoExploration
from ..core.parameters import LearnableParametersDict
from ..core.update import UpdateStrategy
from ..core.warmstart import WarmStartStrategy
from ..optim.gradient_based_optimizer import GradientBasedOptimizer
from .common.agent import ActType, ObsType, SymType
from .common.rl_learning_agent import LrType, RlLearningAgent

# the experience buffer contains the gradient and, possibly, the hessian of the Bellman
# residuals w.r.t. the learnable parameters theta
ExpType: TypeAlias = Union[
    npt.NDArray[np.floating], tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]
]


class LstdQLearningAgent(
    RlLearningAgent[SymType, ExpType, LrType], Generic[SymType, LrType]
):
    """Second-order Least-Squares Temporal Difference (LSTD) Q-learning agent, as first
    proposed in a simpler format in [1], and then in [2].

    The Q-learning agent uses an MPC controller as policy provider and function
    approximation, and adjusts its parametrization according to the temporal-difference
    error, with the goal of improving the policy, in an indirect fashion by learning the
    action value function.

    References
    ----------
    [1] Gros, S. and Zanon, M., 2019. Data-driven economic NMPC using reinforcement
        learning. IEEE Transactions on Automatic Control, 65(2), pp. 636-648.
    [2] Esfahani, H.N., Kordabad, A.B. and Gros, S., 2021, June. Approximate Robust NMPC
        using Reinforcement Learning. In 2021 European Control Conference (ECC), pp.
        132-137. IEEE.
    """

    def __init__(
        self,
        mpc: Mpc[SymType],
        update_strategy: Union[int, UpdateStrategy],
        discount_factor: float,
        optimizer: GradientBasedOptimizer,
        learnable_parameters: LearnableParametersDict[SymType],
        fixed_parameters: Union[
            None, dict[str, npt.ArrayLike], Collection[dict[str, npt.ArrayLike]]
        ] = None,
        exploration: Optional[ExplorationStrategy] = None,
        experience: Union[None, int, ExperienceReplay[ExpType]] = None,
        warmstart: Union[
            Literal["last", "last-successful"], WarmStartStrategy
        ] = "last-successful",
        hessian_type: Literal["none", "approx", "full"] = "approx",
        record_td_errors: bool = False,
        use_last_action_on_fail: bool = False,
        remove_bounds_on_initial_action: bool = False,
        name: Optional[str] = None,
    ) -> None:
        """Instantiates the LSTD Q-learning agent.

        Parameters
        ----------
        mpc : Mpc[casadi.SX or MX]
            The MPC controller used as policy provider by this agent. The instance is
            modified in place to create the approximations of the state function `V(s)`
            and action value function `Q(s,a)`, so it is recommended not to modify it
            further after initialization of the agent. Moreover, some parameter and
            constraint names will need to be created, so an error is thrown if these
            names are already in use in the mpc. These names are under the attributes
            `perturbation_parameter`, `action_parameter` and `action_constraint`.
        update_strategy : UpdateStrategy or int
            The strategy used to decide which frequency to update the mpc parameters
            with. If an `int` is passed, then the default strategy that updates every
            `n` env's steps is used (where `n` is the argument passed); otherwise, an
            instance of `UpdateStrategy` can be passed to specify these in more details.
        discount_factor : float
            In RL, the factor that discounts future rewards in favor of immediate
            rewards. Usually denoted as `\\gamma`. Should be a number in (0, 1].
        optimizer : GradientBasedOptimizer
            A gradient-based optimizer (e.g., `mpcrl.optim.GradientDescent`) to compute
            the updates of the learnable parameters, based on the current gradient-based
            RL algorithm.
        learnable_parameters : LearnableParametersDict
            A special dict containing the learnable parameters of the MPC, together with
            their bounds and values. This dict is complementary with `fixed_parameters`,
            which contains the MPC parameters that are not learnt by the agent.
        fixed_parameters : dict[str, array_like] or collection of, optional
            A dict (or collection of dict, in case of `csnlp.MultistartNlp`) whose keys
            are the names of the MPC parameters and the values are their corresponding
            values. Use this to specify fixed parameters, that is, non-learnable. If
            `None`, then no fixed parameter is assumed.
        exploration : ExplorationStrategy, optional
            Exploration strategy for inducing exploration in the online MPC policy. By
            default `None`, in which case `NoExploration` is used. Should not be set
            when offpolicy learning, as the exploration should be taken care in the
            offpolicy data generation.
        experience : int or ExperienceReplay, optional
            The container for experience replay memory. If `None` is passed, then a
            memory with length 1 is created, i.e., it keeps only the latest memory
            transition.  If an integer `n` is passed, then a memory with the length `n`
            is created and with sample size `n`.
            In the case of LSTD Q-learning, each memory item consists of the action
            value function's gradient and hessian computed at each (succesful) env's
            step.
        warmstart: "last" or "last-successful" or WarmStartStrategy, optional
            The warmstart strategy for the MPC's NLP. If `last-successful`, the last
            successful solution is used to warm start the solver for the next iteration.
            If `last`, the last solution is used, regardless of success or failure.
            Furthermoer, a `WarmStartStrategy` object can be passed to specify a
            strategy for generating multiple warmstart points for the NLP. This is
            useful to generate multiple initial conditions for very non-convex problems.
            Can only be used with an MPC that has an underlying multistart NLP problem
            (see `csnlp.MultistartNlp`).
        hessian_type : {'none', 'approx', 'full'}, optional
            The type of hessian to use in this (potentially) second-order algorithm.
            If 'none', no second order information is used. If `approx`, an easier
            approximation of it is used; otherwise, the full hessian is computed but
            this is much more expensive. This option must be in accordance with the
            choice of `optimizer`, that is, if the optimizer does not use second order
            information, then this option must be set to `none`.
        record_td_errors: bool, optional
            If `True`, the TD errors are recorded in the field `td_errors`, which
            otherwise is `None`. By default, does not record them.
        use_last_action_on_fail : bool, optional
            In case the MPC solver fails
             * if `False`, the action from the last solver's iteration is returned
               anyway (though suboptimal)
             * if `True`, the action from the last successful call to the MPC is
               returned instead (if the MPC has been solved at least once successfully).

            By default, `False`.
        remove_bounds_on_initial_action : bool, optional
            When `True`, the upper and lower bounds on the initial action are removed in
            the action-value function approximator Q(s,a) since the first action is
            constrained to be equal to the initial action. This is useful to avoid
            issues in the LICQ of the NLP. However, it can lead to numerical problems.
            By default, `False`.
        name : str, optional
            Name of the agent. If `None`, one is automatically created from a counter of
            the class' instancies.
        """
        super().__init__(
            mpc=mpc,
            update_strategy=update_strategy,
            discount_factor=discount_factor,
            learnable_parameters=learnable_parameters,
            optimizer=optimizer,
            fixed_parameters=fixed_parameters,
            exploration=exploration,
            experience=experience,
            warmstart=warmstart,
            use_last_action_on_fail=use_last_action_on_fail,
            remove_bounds_on_initial_action=remove_bounds_on_initial_action,
            name=name,
        )
        self.hessian_type = hessian_type
        self._sensitivity = self._init_sensitivity(hessian_type)
        self.td_errors: Optional[list[float]] = [] if record_td_errors else None

    def update(self) -> Optional[str]:
        sample = self.experience.sample()
        if self.hessian_type == "none":
            gradient = np.mean(list(sample), 0)
            return self.optimizer.update(gradient)

        gradients, hessians = zip(*sample)
        gradient = np.mean(gradients, 0)
        hessian = np.mean(hessians, 0)
        return self.optimizer.update(gradient, hessian)

    def train_one_episode(
        self,
        env: Env[ObsType, ActType],
        episode: int,
        init_state: ObsType,
        raises: bool = True,
    ) -> float:
        truncated = terminated = False
        timestep = 0
        rewards = 0.0
        state = init_state
        action_space = getattr(env, "action_space", None)
        # NOTE: if no exploration is set, then we can get away by solving one MPC per
        # iteration, instead of two. The reason is that `a=argmin V(s)` and `min Q(s,a)`
        # are the same thing, because `a` has not been modified by any perturbation.
        no_exploration = isinstance(self.exploration, NoExploration)

        # solve for the first action
        action, solV = self.state_value(state, False, action_space=action_space)
        if not solV.success:
            self.on_mpc_failure(episode, None, solV.status, raises)

        while not (truncated or terminated):
            # compute Q(s,a), or copy V(s) if no exploration is set
            solQ = solV if no_exploration else self.action_value(state, action)

            # step the system with action computed at the previous iteration
            new_state, cost, truncated, terminated, _ = env.step(action)
            self.on_env_step(env, episode, timestep)

            # compute V(s+) and store transition
            new_action, solV = self.state_value(
                new_state, False, action_space=action_space
            )
            if not self._try_store_experience(cost, solQ, solV):
                self.on_mpc_failure(
                    episode, timestep, f"{solQ.status} (Q); {solV.status} (V)", raises
                )

            # increase counters
            state = new_state
            action = new_action
            rewards += float(cost)
            timestep += 1
            self.on_timestep_end(env, episode, timestep)
        return rewards

    def train_one_rollout(
        self,
        rollout: Iterable[tuple[ObsType, ActType, float, ObsType]],
        episode: int,
        raises: bool = True,
    ) -> None:
        # in the case of off-policy q-learning, rollouts are made of
        # State-Action-Reward-next State (SARS) tuples
        for timestep, (state, action, cost, new_state) in enumerate(rollout, start=1):
            # compute Q(s,a)
            solQ = self.action_value(state, action)

            # compute V(s+) and store transition
            _, solV = self.state_value(new_state, False)
            if not self._try_store_experience(cost, solQ, solV):
                self.on_mpc_failure(
                    episode, timestep, f"{solQ.status} (Q); {solV.status} (V)", raises
                )
            self.on_timestep_end("off-policy", episode, timestep)

    def _init_sensitivity(
        self, hessian_type: Literal["none", "approx", "full"]
    ) -> Union[
        Callable[[cs.DM], np.ndarray], Callable[[cs.DM], tuple[np.ndarray, np.ndarray]]
    ]:
        """Internal utility to compute the derivative of Q(s,a) w.r.t. the learnable
        parameters, a.k.a., theta."""
        order = self.optimizer._order
        theta = cs.vvcat(self._learnable_pars.sym.values())
        nlp = (self._V if isinstance(self.exploration, NoExploration) else self._Q).nlp
        x_lam_p = cs.vertcat(nlp.primal_dual, nlp.p)

        # compute first order sensitivity
        snlp = NlpSensitivity(nlp, theta)
        gradient = snlp.jacobians["L-p"]  # exact gradient, i.e., dQ/dtheta

        if hessian_type == "none":
            assert order == 1, "Expected 1st-order optimizer with `hessian_type=none`."

            sensitivity = cs.Function(
                "S", (x_lam_p,), (gradient,), ("x_lam_p",), ("dQ",), {"cse": True}
            )

            def func(sol_values: cs.DM) -> np.ndarray:
                return np.asarray(sensitivity(sol_values).elements())

        else:
            assert (
                order == 2
            ), "Expected 2nd-order optimizer with `hessian_type=approx` or `full`."

            # compute second order sensitivity
            hessian = snlp.hessians["L-pp"]  # approximate hessian
            if hessian_type == "full":
                Kp = snlp.jacobians["K-p"]
                Ky = snlp.jacobians["K-y"]
                dydtheta = -cs.solve(Ky, Kp)
                Lpy = cs.jacobian(gradient, nlp.primal_dual)
                hessian += Lpy @ dydtheta  # not sure if Lpy or Kp.T

            sensitivity = cs.Function(
                "S",
                (x_lam_p,),
                (gradient, hessian),
                ("x_lam_p",),
                ("dQ", "ddQ"),
                {"cse": True},
            )

            def func(sol_values: cs.DM) -> tuple[np.ndarray, np.ndarray]:
                dQ, ddQ = sensitivity(sol_values)
                return np.asarray(dQ.elements()), ddQ.toarray()

        return func

    def _try_store_experience(
        self, cost: SupportsFloat, solQ: Solution[SymType], solV: Solution[SymType]
    ) -> bool:
        """Internal utility that tries to store the gradient and hessian for the current
        transition in memory, if both V and Q were successful; otherwise, does not store
        it. Returns whether it was successful or not."""
        success = solQ.success and solV.success
        if success:
            sol_values = solQ.all_vals
            td_error = cost + self.discount_factor * solV.f - solQ.f
            if self.hessian_type == "none":
                dQ = self._sensitivity(sol_values)
                gradient = -td_error * dQ
                self.store_experience(gradient)
            else:
                dQ, ddQ = self._sensitivity(sol_values)
                gradient = -td_error * dQ
                hessian = np.multiply.outer(dQ, dQ) - td_error * ddQ
                self.store_experience((gradient, hessian))
        else:
            td_error = np.nan

        if self.td_errors is not None:
            self.td_errors.append(td_error)
        return success



from typing import Literal, Optional, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.schedulers import Scheduler
from .gradient_based_optimizer import GradientBasedOptimizer, LrType, SymType


class GradientDescent(GradientBasedOptimizer[SymType, LrType]):
    """First-order Gradient descent optimizer, based on [1,2].

    References
    ----------
    [1] Sutskever, I., Martens, J., Dahl, G. and Hinton, G., 2013, May. On the
        importance of initialization and momentum in deep learning. In International
        conference on machine learning (pp. 1139-1147). PMLR.
    [2] SGD - PyTorch 2.1 documentation.
        https://pytorch.org/docs/stable/generated/torch.optim.SGD.html
    """

    _order = 1
    _hessian_sparsity = "diag"
    """In GD, the hessian is at most diagonal, i.e., in case we have constraints."""

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        weight_decay: float = 0.0,
        momentum: float = 0.0,
        dampening: float = 0.0,
        nesterov: bool = False,
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
    ) -> None:
        """Instantiates the optimizer.

        Parameters
        ----------
        learning_rate : float/array, scheduler
            The learning rate of the optimizer. A float/array can be passed in case the
            learning rate must stay constant; otherwise, a scheduler can be passed which
            will be stepped `on_update` by default (see `hook` argument).
        weight_decay : float, optional
            A positive float that specifies the decay of the learnable parameters in the
            form of an L2 regularization term. By default, it is set to `0.0`, so no
            decay/regularization takes place.
        momentum : float, optional
            A positive float that specifies the momentum factor. By default, it is set
            to `0.0`, so no momentum is used.
        dampening : float, optional
            A positive float that specifies the dampening factor for the momentum. By
            default, it is set to `0.0`, so no dampening is used.
        nesterov : bool, optional
            A boolean that specifies whether to use Nesterov momentum. By default, it is
            set to `False`.
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies when to step the optimizer's learning rate's scheduler to decay
            its value (see `step` method also). This allows to vary the rate over the
            learning iterations. The options are:
             - `on_update` steps the learning rate after each agent's update
             - `on_episode_end` steps the learning rate after each episode's end
             - `on_timestep_end` steps the learning rate after each env's timestep.

            By default, 'on_update' is selected.
        max_percentage_update : float, optional
            A positive float that specifies the maximum percentage change the learnable
            parameters can experience in each update. For example,
            `max_percentage_update=0.5` means that the parameters can be updated by up
            to 50% of their current value. By default, it is set to `+inf`. If
            specified, the update becomes constrained and has to be solved as a QP,
            which is inevitably slower than its unconstrained counterpart.
        """
        super().__init__(learning_rate, hook, max_percentage_update)
        self.weight_decay = weight_decay
        self.momentum = momentum
        self.dampening = dampening
        self.nesterov = nesterov
        self._momentum_buffer = None

    def _first_order_update(
        self, gradient: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        theta = self.learnable_parameters.value

        # compute candidate update
        dtheta, self._momentum_buffer = _gd(
            theta,
            gradient,
            self._momentum_buffer,
            self.weight_decay,
            self.momentum,
            self.lr_scheduler.value,
            self.dampening,
            self.nesterov,
        )

        # if unconstrained, apply the update directly; otherwise, solve the QP
        solver = self._update_solver
        if solver is None:
            return theta + dtheta, None
        lbx, ubx = self._get_update_bounds(theta)
        sol = solver(h=cs.DM.eye(theta.shape[0]), g=-dtheta, lbx=lbx, ubx=ubx)
        dtheta = np.asarray(sol["x"].elements())
        stats = solver.stats()
        return theta + dtheta, None if stats["success"] else stats["return_status"]


def _gd(
    theta: np.ndarray,
    g: np.ndarray,
    momentum_buffer: Optional[np.ndarray],
    weight_decay: float,
    momentum: float,
    lr: LrType,
    dampening: float,
    nesterov: bool,
) -> tuple[np.ndarray, Optional[np.ndarray]]:
    """Computes the update's change according to the gradient descent algorithm."""
    if weight_decay > 0.0:
        g += weight_decay * theta
    if momentum > 0.0:
        if momentum_buffer is None:
            momentum_buffer = g.copy()
        else:
            momentum_buffer = momentum * momentum_buffer + (1 - dampening) * g
        if nesterov:
            g += momentum * momentum_buffer
        else:
            g = momentum_buffer
    dtheta = -lr * g
    return dtheta, momentum_buffer



from typing import Any, Literal, Optional, TypeVar

import numpy as np
import numpy.typing as npt
from gymnasium import Env

from ...agents.common.learning_agent import ExpType, LearningAgent
from ...util.iters import bool_cycle
from ...util.seeding import RngType, mk_seed
from .wrapper import LearningWrapper, SymType

ObsType = TypeVar("ObsType")
ActType = TypeVar("ActType")


class Evaluate(LearningWrapper[SymType, ExpType]):
    """Wrapper for evaluating an agent during learning."""

    def __init__(
        self,
        agent: LearningAgent[SymType, ExpType],
        eval_env: Env[ObsType, ActType],
        hook: Literal["on_episode_end", "on_timestep_end", "on_update"],
        frequency: int,
        n_eval_episodes: int = 1,
        eval_immediately: bool = False,
        *,
        deterministic: bool = True,
        seed: RngType = None,
        raises: bool = True,
        env_reset_options: Optional[dict[str, Any]] = None,
        fix_seed: bool = False,
    ) -> None:
        """Creates an instance of the evaluation wrapper around the agent.

        Parameters
        ----------
        agent : LearningAgent
            The learning agent to be evaluated by the wrapper.
        eval_env : gymnasium.Env
            A gym environment to evaluate the agent in.
        hook : {'on_episode_end', 'on_timestep_end', 'on_update'}
            Hook to trigger the evaluation. The evaluation will be triggered every
            `frequency` invokations of the specified hook.
        frequency : int
            Frequency of the evaluation.
        n_eval_episodes : int, optional
            How many episodes to evaluate the agent for, by default `1`.
        eval_immediately : bool, optional
            Whether to evaluate the agent immediately after the wrapper is created, by
            default `False`.
        deterministic : bool, optional
            Whether the agent should act deterministically; by default, `True`.
        seed : None, int, array_like[ints], SeedSequence, BitGenerator, Generator
            Agent's and each env's RNG seed for the evaluation.
        raises : bool, optional
            If `True`, when any of the MPC solver runs fails, or when an update fails,
            the corresponding error is raised; otherwise, only a warning is raised.
        env_reset_options : dict, optional
            Additional information to specify how the environment is reset at each
            evalution episode (optional, depending on the specific environment).
        fix_seed : bool, optional
            If `True`, the seed is fixed and the same seed is used for all evaluations.
        """
        self.eval_env = eval_env
        self._hook = hook
        self._n_eval_episodes = n_eval_episodes
        self._deterministic = deterministic
        np_random = np.random.default_rng(seed)
        self._seed = mk_seed(np_random) if fix_seed else np_random
        self._raises = raises
        self._env_reset_options = env_reset_options
        self._keep_seed_fixed = fix_seed
        self._eval_cycle = bool_cycle(frequency)
        self.eval_returns: list[npt.NDArray[np.floating]] = []
        self._is_eval_in_progress = False
        super().__init__(agent)
        if eval_immediately:
            self._evaluate(force=True)

    def _evaluate(self, *_: Any, **__: Any) -> None:
        """Internal utility to perform the evaluation.

        Note: we do not check if the agent is training or evaluating, so the wrapper's
        evaluation round can be initiated in both cases, the second making less sense.
        """
        if self._is_eval_in_progress or (
            not __.get("force", False) and not next(self._eval_cycle)
        ):
            return
        self._is_eval_in_progress = True
        unwrapped_agent = self.agent.unwrapped
        updates_flag = unwrapped_agent._updates_enabled
        try:
            self.eval_returns.append(
                self.agent.evaluate(
                    self.eval_env,
                    self._n_eval_episodes,
                    self._deterministic,
                    self._seed,
                    self._raises,
                    self._env_reset_options,
                )
            )
        finally:
            self._is_eval_in_progress = False
            unwrapped_agent._updates_enabled = updates_flag

    def establish_callback_hooks(self) -> None:
        super().establish_callback_hooks()
        self.hook_callback(repr(self), self._hook, self._evaluate)



from collections.abc import Generator
from itertools import chain
from typing import Literal, Optional

import numpy as np
from csnlp.multistart import RandomStartPoints, StructuredStartPoints
from numpy import typing as npt

from ..util.seeding import RngType


class WarmStartStrategy:
    """Class containing all the information to guide the warm start strategy for the
    MPC's NLP in order to speed up computations (by selecting appropriate initial
    conditions) and to support multiple initial conditions."""

    def __init__(
        self,
        warmstart: Literal["last", "last-successful"] = "last-successful",
        structured_points: Optional[StructuredStartPoints] = None,
        random_points: Optional[RandomStartPoints] = None,
        update_biases_for_random_points: bool = True,
        seed: RngType = None,
    ) -> None:
        """Instantiates a warm start strategy for solving the MPC's NLP.

        Parameters
        ----------
        warmstart: "last" or "last-successful", optional
            The warmstart strategy for the MPC's NLP. If `last-successful`, the last
            successful solution is used to warm start the solver for the next iteration.
            If `last`, the last solution is used, regardless of success or failure.
        structured_points : StructuredStartPoint, optional
            Class containing info on how to generate structured starting points for the
            NLP. If `None`, no structured point is generated.
        random_points : RandomStartPoints, optional
            Class containing info on how to generate random starting points for the NLP.
            Its `biases` field will be automatically updated with the last successful
            solution, unless disabled with `update_biases_for_random_points=False`). If
            `None`, no random point is generated.
        update_biases_for_random_points : bool, optional
            If `True`, the random points are biased around the bias-values updated with
            the latest successful MPC solution. If `False`, the biases in
            `random_points` are not updated.
        seed : None, int, array of ints, SeedSequence, BitGenerator, Generator
            Seed for the random number generator. By default, `None`.
        """
        self.store_always = warmstart == "last"
        self.structured_points = structured_points
        self.random_points = random_points
        self.update_biases_for_random_points = update_biases_for_random_points
        self.reset(seed)

    @property
    def n_points(self) -> int:
        """Returns the number of both random and structured starting points."""
        return (0 if self.random_points is None else self.random_points.multistarts) + (
            0 if self.structured_points is None else self.structured_points.multistarts
        )

    def reset(self, seed: RngType = None) -> None:
        """Resets the sampling RNG."""
        if self.random_points is not None:
            self.random_points.np_random = np.random.default_rng(seed)

    def generate(
        self, biases: Optional[dict[str, npt.ArrayLike]] = None
    ) -> Generator[dict[str, npt.ArrayLike], None, None]:
        """Generates initial conditions for the MPC's NLP.

        Parameters
        ----------
        biases : dict of (str, array_like), optional
            Optional biases that can be used to update the random points' original
            biases. If `None` or `update_biases_for_random_points=False`, the original
            biases are kept constant. These do not affect the generation of structure
            points in any way.

        Yields
        ------
        Generator of dict of (str, array_like)
            Yields the initial conditions for the MPC's NLP.
        """
        to_be_chained = []

        if self.structured_points is not None:
            to_be_chained.append(self.structured_points)

        if self.random_points is not None:
            if self.update_biases_for_random_points and biases is not None:
                self.random_points.biases.update(biases)
            to_be_chained.append(self.random_points)

        return chain.from_iterable(to_be_chained)

    def __repr__(self) -> str:
        nr = 0 if self.random_points is None else self.random_points.multistarts
        ns = 0 if self.structured_points is None else self.structured_points.multistarts
        return (
            f"{self.__class__.__name__}(store_always={self.store_always},"
            f"random_points={nr},structured_points={ns})"
        )

    def __str__(self) -> str:
        return self.__repr__()



from typing import Generic, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.schedulers import NoScheduling, Scheduler
from .base_optimizer import BaseOptimizer, SymType

LrType = TypeVar("LrType", npt.NDArray[np.floating], float)


class GradientBasedOptimizer(BaseOptimizer[SymType], Generic[SymType, LrType]):
    """Base class for first- and second-order gradient-based optimization algorithms."""

    _order: Literal[1, 2]
    """Order of the optimizer: 1 for first-order, 2 for second-order."""

    _hessian_sparsity: Literal["dense", "diag"]
    """Sparsity of the hessian. It can be overridden by each subclass, in case a
    particular structure is known, e.g., diagonal."""

    def __init__(
        self,
        learning_rate: Union[LrType, Scheduler[LrType]],
        hook: Literal["on_update", "on_episode_end", "on_timestep_end"] = "on_update",
        max_percentage_update: float = float("+inf"),
    ) -> None:
        """Instantiates the optimizer.

        Parameters
        ----------
        learning_rate : float/array, scheduler
            The learning rate of the optimizer. A float/array can be passed in case the
            learning rate must stay constant; otherwise, a scheduler can be passed which
            will be stepped `on_update` by default (see `hook` argument).
        hook : {'on_update', 'on_episode_end', 'on_timestep_end'}, optional
            Specifies when to step the optimizer's learning rate's scheduler to decay
            its value (see `step` method also). This allows to vary the rate over the
            learning iterations. The options are:
             - `on_update` steps the learning rate after each agent's update
             - `on_episode_end` steps the learning rate after each episode's end
             - `on_timestep_end` steps the learning rate after each env's timestep.

            By default, 'on_update' is selected.
        max_percentage_update : float, optional
            A positive float that specifies the maximum percentage change the learnable
            parameters can experience in each update. For example,
            `max_percentage_update=0.5` means that the parameters can be updated by up
            to 50% of their current value. By default, it is set to `+inf`.
        """
        super().__init__(max_percentage_update)
        if not isinstance(learning_rate, Scheduler):
            learning_rate = NoScheduling[LrType](learning_rate)
        self.lr_scheduler: Scheduler[LrType] = learning_rate
        self._hook = hook
        self._update_solver: cs.Function

    @property
    def hook(self) -> Optional[str]:
        """Specifies to which callback to hook, i.e., when to step the learning rate's
        scheduler to decay its value (see `step` method also). Can be `None` in case no
        hook is needed."""
        # return hook only if the learning rate scheduler requires to be stepped
        return None if isinstance(self.lr_scheduler, NoScheduling) else self._hook

    def step(self, *_, **__) -> None:
        """Steps/decays the learning rate according to its scheduler."""
        self.lr_scheduler.step()

    def _init_update_solver(self) -> Optional[cs.Function]:
        """Internal utility to initialize, if the learnable parameters are constrained,
        a constrained update solver which, by default, is a QP. Otherwise, no solver is
        required to perform the update.
        """
        if (
            self.max_percentage_update == float("+inf")
            and np.isneginf(self.learnable_parameters.lb).all()
            and np.isposinf(self.learnable_parameters.ub).all()
        ):
            return None
        n_params = self.learnable_parameters.size
        qp = {"h": getattr(cs.Sparsity, self._hessian_sparsity)(n_params, n_params)}
        opts = {
            "error_on_fail": False,
            "osqp": {
                "verbose": False,
                "polish": True,
                "scaling": 20,
                "eps_abs": 1e-9,
                "eps_rel": 1e-9,
                "eps_prim_inf": 1e-10,
                "eps_dual_inf": 1e-10,
                "max_iter": 6000,
            },
        }
        return cs.conic(f"qpsol_{id(self)}", "osqp", qp, opts)

    def update(
        self,
        gradient: npt.NDArray[np.floating],
        hessian: Optional[npt.NDArray[np.floating]] = None,
    ) -> Optional[str]:
        """Computes the gradient update of the learnable parameters dictated by the
        current RL algorithm.

        Parameters
        ----------
        gradient : 1D array
            The gradient of the learnable parameters.
        hessian : None, or 2D array
            The hessian of the learnable parameters. When the optimizer is firt-order,
            it is expected to be `None` (as it is unused). When the optimizer is
            second-order, it is expected to be a 2D array.

        Returns
        -------
        status : str, optional
            An optional string containing the status of the update, e.g., the status of
            the QP solver, if used.
        """
        if self._order == 1:
            theta_new, status = self._first_order_update(gradient)
        else:
            theta_new, status = self._second_order_update(gradient, hessian)
        # theta_new = np.clip(
        #     theta_new, self.learnable_parameters.lb, self.learnable_parameters.ub
        # )
        self.learnable_parameters.update_values(theta_new)
        return status

    def _first_order_update(
        self, gradient: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        """Internally runs a first order update."""
        raise NotImplementedError(
            f"`{self.__class__.__name__}` optimizer does not implement "
            "`_first_order_update`"
        )

    def _second_order_update(
        self, gradient: npt.NDArray[np.floating], hessian: npt.NDArray[np.floating]
    ) -> tuple[npt.NDArray[np.floating], Optional[str]]:
        """Internally runs a second order update."""
        raise NotImplementedError(
            f"`{self.__class__.__name__}` optimizer does not implement "
            "`_second_order_update`"
        )

    def __repr__(self) -> str:
        cn = self.__class__.__name__
        hookstr = "None" if self.hook is None else f"'{self.hook}'"
        mp = self.max_percentage_update
        return f"{cn}(lr={self.lr_scheduler},hook={hookstr},max%={mp})"



from collections.abc import Iterator
from itertools import count
from typing import Optional


class Named:
    """Base class for objects with names."""

    __ids: dict[type, Iterator[int]] = {}

    def __init__(self, name: Optional[str] = None) -> None:
        """Instantiates the class with the given `name` attribute.

        Parameters
        ----------
        name : str, optional
            Name of the object. If `None`, one is automatically created from a counter
            of the class' instancies.
        """
        cls = self.__class__
        if cls in self.__ids:
            _id = self.__ids[cls]
        else:
            _id = count(0)
            self.__ids[cls] = _id
        self.name = name or f"{cls.__name__}{next(_id)}"

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name}>"
